#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#ifndef __USE_XOPEN
#define __USE_XOPEN
#endif
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <fnmatch.h>
#include <errno.h>
#include "dftpd.h"
#undef ERR
#include <curses.h>

/* libdftp.c dependencies */

char *dftp_identip2text(struct ident_ip *, int);
char *dftp_flags2text(int, int);

int dftp_ugetent(int, struct dftp_uent *);
int dftp_uaddent(int, struct dftp_uent *);
int dftp_uaddsiteent(int, struct dftp_uent *);
void dftp_uupdatedb(int, struct dftp_uent *);
void dftp_uupdatesitedb(int, int);
int dftp_ugetbyname(int, char *, struct dftp_uent *);
void dftp_ufree(struct dftp_uent *);

int dftp_ggetent(int, struct dftp_gent *);
int dftp_gaddent(int, struct dftp_gent *);
int dftp_gaddsiteent(int, struct dftp_gent *);
void dftp_gupdatedb(int, struct dftp_gent *);
void dftp_gupdatesitedb(int, int);
int dftp_ggetbyid(int, int, struct dftp_gent *);
int dftp_ggetbyname(int, char *, struct dftp_gent *);
void dftp_gfree(struct dftp_gent *);

void dftp_readconfig(struct dftp_config *);
char *dftp_printbytes(unsigned long long);

struct dftp_config cfg;
char *passwdname;
char *sitepasswdname;
char *groupname;
char *sitegroupname;

struct dftp_uent **ulist = NULL;
struct dftp_gent **glist = NULL;
unsigned int unum = 0, gnum = 0, visual = 0, mainwinw, mainwinh;
WINDOW *mainwin = NULL;
const char titletext[] = "Demi-FTPd configuration utility v1.2";
const char bottomtext[] = "Press F1 or h for help";

#define USERWINW 62
#define USERWINH 20
#define USERWINOFF 28
#define GROUPWINW 60
#define GROUPWINH 10
#define GROUPWINOFF 26
#define USERFIELDS 12
#define GROUPFIELDS 2
#define MENUITEMS 4
#define MENUW 30
#define MENUH MENUITEMS+2

void drawwin(WINDOW *win, int w, int h, const char *title, const char *bottom)
{
   int i;

   wmove(win,0,0);
   for(i=0;i<w*h;i++) waddch(win,' ');
   wattroff(win,A_REVERSE);
   if(has_colors()) wborder(win,0,0,0,0,0,0,0,0);
   else wborder(win,'|','|','-','-','+','+','+','+');
   if(title)
   {
      wmove(win,0,w/2-strlen(title)/2-1);
      if(has_colors()) waddch(win,ACS_RTEE); else waddch(win,'|');
      waddstr(win,title);
      if(has_colors()) waddch(win,ACS_LTEE); else waddch(win,'|');
   }
   if(bottom)
   {
      wmove(win,h-1,w/2-strlen(bottom)/2-1);
      if(has_colors()) waddch(win,ACS_RTEE); else waddch(win,'|');
      waddstr(win,bottom);
      if(has_colors()) waddch(win,ACS_LTEE); else waddch(win,'|');
   }
   wrefresh(win);
}

void init_mainwin(int w, int h)
{
   mainwin = newwin(h,w,LINES/2-h/2,COLS/2-w/2);
   keypad(mainwin,TRUE); wattrset(mainwin,COLOR_PAIR(1));
   mainwinw = w; mainwinh = h;
   drawwin(mainwin,mainwinw,mainwinh,titletext,bottomtext);
}

void notify(char *format, ...)
{
   WINDOW *notifywin; char text[120]; va_list ap; int w;

   va_start(ap,format); vsprintf(text,format,ap);
   w = strlen(text)+4;
   notifywin = newwin(3,w,LINES/2-2,COLS/2-w/2);
   keypad(notifywin,TRUE); wattrset(notifywin,COLOR_PAIR(3));
   drawwin(notifywin,w,3,NULL,NULL); mvwaddstr(notifywin,1,2,text);
   while(wgetch(notifywin) != 10);
   werase(notifywin); wrefresh(notifywin); delwin(notifywin);
   if(mainwin)
   {
      touchwin(mainwin); wrefresh(mainwin);
   }
   va_end(ap);
}

int confirm(char *text, int ydefault)
{
   WINDOW *confirmwin; int w = strlen(text)+10;
   char yntext[7];

   if(ydefault) strcpy(yntext," [Y/n]");
   else strcpy(yntext," [y/N]");
   confirmwin = newwin(3,w,LINES/2-2,COLS/2-w/2);
   keypad(confirmwin,TRUE); wattrset(confirmwin,COLOR_PAIR(3));
   drawwin(confirmwin,w,3,NULL,NULL);
   mvwaddstr(confirmwin,1,2,text); waddstr(confirmwin,yntext);
   wrefresh(confirmwin);
   do
   {
      w = wgetch(confirmwin);
   } while(w != 'y' && w != 'n' && w != 10);
   werase(confirmwin); wrefresh(confirmwin); delwin(confirmwin);
   if(mainwin)
   {
      touchwin(mainwin); wrefresh(mainwin);
   }
   switch(w)
   {
      case 'y': return 1;
      case 'n': return 0;
   }
   return ydefault;
}

void visual_setup()
{
   if(initscr() == NULL)
   {
      fprintf(stderr,"Error initializing ncurses\n");
      exit(1);
   }
   if(LINES < USERWINH || COLS < USERWINW)
   {
      fprintf(stderr,"Your terminal screen is too small - at least %ix%i is required.\n",USERWINW,USERWINH);
      endwin(); exit(2);
   }

   cbreak(); noecho();
   curs_set(0);
   if(has_colors())
   {
      start_color();
      init_pair(1,COLOR_WHITE,COLOR_BLUE);
      init_pair(2,COLOR_WHITE,COLOR_BLACK);
      init_pair(3,COLOR_WHITE,COLOR_RED);
      init_pair(4,COLOR_WHITE,COLOR_MAGENTA);
      init_pair(5,COLOR_WHITE,COLOR_GREEN);
   }
   visual = 1;
}

int usage()
{
   printf("%s\n",titletext);
   printf("Usage: dftpcfg [command]\n\n");
   printf("                            rebuild   rebuild site's passwd and group files\n");
   printf("                       ulist [mask]   list either all or mask-specified users\n");
   printf("                       glist [mask]   list either all or mask-specified groups\n");
   exit(0);
}

void listusers(char *args)
{
   int i,pwdfile; struct dftp_uent sala; char *teksti = NULL;

   pwdfile = open(passwdname,O_RDONLY);
   if(pwdfile < 0)
   {
      fprintf(stderr,"Error accessing user database: %s\n",strerror(errno));
      return;
   }

   if(args == NULL)
   {
      args = malloc(2); strcpy(args,"*");
   }
   while( (i = dftp_ugetent(pwdfile, &sala)) )
   {
      if(i < 0 && !fnmatch(args,sala.name,FNM_NOESCAPE|FNM_PERIOD))
      {
         printf("Name:%-16s ",sala.name); 
         printf("Uid:%i ",sala.uid); 
         printf("Gid:%i Ratio:%i ",sala.gid,sala.ratio);
         printf("Flags:%s\n",sala.flags);
         printf("Max connections:");
         if(!sala.maxconn) printf("- UL:");
         else printf("%i UL:",sala.maxconn);
         printf("%s DL:",dftp_printbytes(sala.ulbytes));
         printf("%s Credits:",dftp_printbytes(sala.dlbytes));
         printf("%s\n",dftp_printbytes(sala.credits));
         teksti = dftp_identip2text(sala.identip,sala.identipcount);
         printf("Ident/IP:%s\n",teksti); free(teksti);
      }
      dftp_ufree(&sala);
   }
   close(pwdfile);
}

void listgroups(char *args)
{
   int i,grpfile; struct dftp_gent sala;

   grpfile = open(groupname,O_RDONLY);
   if(grpfile < 0)
   {
      fprintf(stderr,"Error accessing group database: %s\n",strerror(errno));
      return;
   }

   if(args == NULL)
   {
      args = malloc(2); strcpy(args,"*");
   }
   while( (i = dftp_ggetent(grpfile, &sala)) )
   {
      if(i < 0 && !fnmatch(args,sala.name,FNM_NOESCAPE|FNM_PERIOD))
      {
         printf("Name:%-16s Id:%i\n",sala.name,sala.id); 
         printf("Members:");
         for(i=0;i<sala.membercount;i++)
         {
            printf("%s",sala.members[i]);
            if(i != sala.membercount-1) printf(",");
         }
         printf("\n");
      }
      dftp_gfree(&sala);
   }
   close(grpfile);
}

void rebuild()
{
   int pwdfile = -1, grpfile = -1, pwdsitefile = -1, grpsitefile = -1;

   if(!visual)
   {
      printf("Rebuilding %s...",sitepasswdname); fflush(stdout);
   }
   pwdfile = open(passwdname,O_RDONLY);
   if(pwdfile < 0)
   {
      fprintf(stderr,"Error opening %s for read/write: %s\n",passwdname,strerror(errno));
      return;
   }
   pwdsitefile = open(sitepasswdname,O_CREAT|O_RDWR,0644);
   if(pwdsitefile < 0)
   {
      fprintf(stderr,"Error opening %s for read/write: %s\n",sitepasswdname,strerror(errno));
      close(pwdfile); return;
   }
   flock(pwdfile,LOCK_EX); flock(pwdsitefile,LOCK_EX);
   dftp_uupdatesitedb(pwdfile,pwdsitefile);
   flock(pwdfile,LOCK_UN); flock(pwdsitefile,LOCK_UN);
   if(!visual) printf("done\n");

   if(!visual)
   {
      printf("Rebuilding %s...",sitegroupname); fflush(stdout);
   }
   grpfile = open(groupname,O_RDONLY);
   if(grpfile < 0)
   {
      fprintf(stderr,"Error opening %s for read/write: %s\n",groupname,strerror(errno));
      return;
   }
   grpsitefile = open(sitegroupname,O_CREAT|O_RDWR,0644);
   if(grpsitefile < 0)
   {
      fprintf(stderr,"Error opening %s for read/write: %s\n",sitegroupname,strerror(errno));
      close(grpfile); return;
   }
   flock(grpfile,LOCK_EX); flock(grpsitefile,LOCK_EX);
   dftp_gupdatesitedb(grpfile,grpsitefile);
   flock(grpfile,LOCK_UN); flock(grpsitefile,LOCK_UN);
   if(!visual) printf("done\n");
}

int load_users()
{
   int i,fd; struct dftp_uent entry;

   fd = open(passwdname,O_RDONLY);
   if(fd < 0) return 0;
   flock(fd,LOCK_EX);
   while((i=dftp_ugetent(fd,&entry)))
      if(i < 0)
      {
         ulist = realloc(ulist,(unum+1)*sizeof(void *));
         ulist[unum] = malloc(sizeof(struct dftp_uent));
         memcpy(ulist[unum++],&entry,sizeof(struct dftp_uent));
      }
   flock(fd,LOCK_UN);
   close(fd); return 1;
}

int save_users()
{
   int i,fd;

   fd = open(passwdname,O_RDWR|O_CREAT,0644);
   if(fd < 0) return 0;
   flock(fd,LOCK_EX);
   for(i=0;i<unum;i++) dftp_uaddent(fd,ulist[i]);
   ftruncate(fd,lseek(fd,0,SEEK_CUR));
   flock(fd,LOCK_UN);
   close(fd);

   fd = open(sitepasswdname,O_RDWR|O_CREAT,0644);
   if(fd < 0) return 0;
   flock(fd,LOCK_EX);
   write(fd,"root:*:0:0:::\n",strlen("root:*:0:0:::\n"));
   for(i=0;i<unum;i++) dftp_uaddsiteent(fd,ulist[i]);
   ftruncate(fd,lseek(fd,0,SEEK_CUR));
   flock(fd,LOCK_UN);
   close(fd);
   return 1;
}

int load_groups()
{
   int i,fd; struct dftp_gent entry;

   fd = open(groupname,O_RDONLY);
   if(fd < 0) return 0;
   flock(fd,LOCK_EX);
   while((i=dftp_ggetent(fd,&entry)))
      if(i < 0)
      {
         glist = realloc(glist,(gnum+1)*sizeof(void *));
         glist[gnum] = malloc(sizeof(struct dftp_gent));
         memcpy(glist[gnum++],&entry,sizeof(struct dftp_gent));
      }
   flock(fd,LOCK_UN);
   close(fd); return 1;
}

int save_groups()
{
   int i,fd;

   fd = open(groupname,O_RDWR|O_CREAT,0644);
   if(fd < 0) return 0;
   flock(fd,LOCK_EX);
   for(i=0;i<gnum;i++) dftp_gaddent(fd,glist[i]);
   ftruncate(fd,lseek(fd,0,SEEK_CUR));
   flock(fd,LOCK_UN);
   close(fd);

   fd = open(sitegroupname,O_RDWR|O_CREAT,0644);
   if(fd < 0) return 0;
   flock(fd,LOCK_EX);
   write(fd,"root::0:root\n",strlen("root::0:root\n"));
   for(i=0;i<gnum;i++) dftp_gaddsiteent(fd,glist[i]);
   ftruncate(fd,lseek(fd,0,SEEK_CUR));
   flock(fd,LOCK_UN);
   close(fd);
   return 1;
}

void free_entries()
{
   while(unum)
   {
      dftp_ufree(ulist[unum-1]); free(ulist[--unum]);
   }
   free(ulist); ulist = NULL; unum = 0;
   while(gnum)
   {
      dftp_gfree(glist[gnum-1]); free(glist[--gnum]);
   }
   free(glist); glist = NULL; gnum = 0;
}

void menuitem(WINDOW *akkuna,char *item,int posy,int hlpos)
{
   int posx;

   if(hlpos == posy) wattron(akkuna,A_REVERSE);
   wmove(akkuna,posy+1,1);
   for(posx=1;posx<MENUW-1;posx++) waddch(akkuna,' ');
   posx = MENUW/2-strlen(item)/2;
   mvwaddstr(akkuna,posy+1,posx,item);
   if(hlpos == posy) wattroff(akkuna,A_REVERSE);
}

void addmember(struct dftp_gent *group, char *name)
{
   int i;

   for(i=0;i<group->membercount;i++)
      if(!strcmp(group->members[i],name)) break;
   if(i == group->membercount)
   {
      group->members = realloc(group->members,
       ++group->membercount*sizeof(void *));
      group->members[group->membercount-1] = malloc(strlen(name)+1);
      strcpy(group->members[group->membercount-1],name);
   }
}

int validateusers()
{
   int i, usernum;

   for(usernum=0;usernum<unum;usernum++)
   {
      if(!strlen(ulist[usernum]->name))
      {
         notify("Error - user #%i has no name",usernum+1); return 0;
      }
      if(isdigit(ulist[usernum]->name[0]))
      {
         notify("Error - first letter in name of user #%i is a number",usernum+1);
         return 0;
      }
      for(i = 0; i < strlen(ulist[usernum]->name); i++)
      {
         if(!isalnum(ulist[usernum]->name[i]))
         {
            if(ulist[usernum]->name[i] == '\n') ulist[usernum]->name[i] = '\0';
            else
            {
               notify("Error - name of user #%i contains non-alphanumeric characters",usernum+1);
               return 0;
            }
         }
      }

      for(i = usernum+1; i < unum; i++)
      {
         if(ulist[usernum]->uid == ulist[i]->uid)
         {
            notify("Error - users #%i and #%i have duplicate user IDs",usernum+1,i+1);
            return 0;
         }
         if(!strcmp(ulist[usernum]->name,ulist[i]->name))
         {
            notify("Error - users #%i and #%i have duplicate names",usernum+1,i+1);
            return 0;
         }
      }
      for(i = 0; i < gnum; i++) if(glist[i]->id == ulist[usernum]->gid) break;
      if(i == gnum)
      {
         notify("Error - primary group for user #%i does not exist",usernum+1);
         return 0;
      }
      else addmember(glist[i],ulist[usernum]->name);
   }
   return 1;
}

int validategroups()
{
   int i, groupnum;

   for(groupnum=0;groupnum<gnum;groupnum++)
   {
      if(!strlen(glist[groupnum]->name))
      {
         notify("Error - group #%i has no name",groupnum+1); return 0;
      }
      if(isdigit(glist[groupnum]->name[0]))
      {
         notify("Error - first letter in name of group #%i is a number",groupnum+1);
         return 0;
      }
      for(i = 0; i < strlen(glist[groupnum]->name); i++)
      {
         if(!isalnum(glist[groupnum]->name[i]))
         {
            if(glist[groupnum]->name[i] == '\n') glist[groupnum]->name[i] = '\0';
            else
            {
               notify("Error - name of group #%i contains non-alphanumeric characters",groupnum+1);
               return 0;
            }
         }
      }
      for(i = groupnum+1; i < gnum; i++)
      {
         if(glist[groupnum]->id == glist[i]->id)
         {
            notify("Error - groups #%i and #%i have duplicate IDs",groupnum+1,i+1);
            return 0;
         }
         if(!strcmp(glist[groupnum]->name,glist[i]->name))
         {
            notify("Error - groups #%i and #%i have duplicate names",groupnum+1,i+1);
            return 0;
         }
      }
   }
   return 1;
}

void printuser(int usernum, int fieldnum)
{
   int i,j;

   for(i=2;i<USERWINH-2;i++)
   {
      wmove(mainwin,i,2);
      for(j=0;j<USERWINW-4;j++) waddch(mainwin,' ');
   }
   if(unum) mvwprintw(mainwin,2,2,"User %i of %i",usernum,unum);
   else
   {
      mvwprintw(mainwin,2,2,"User 0 of 0"); return;
   }
   if(fieldnum == 1) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,5,2,"Name");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD); wprintw(mainwin,ulist[usernum-1]->name);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 2) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,7,2,"UID");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD); wprintw(mainwin,"%-5hu ",ulist[usernum-1]->uid);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 3) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,9,2,"GID");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD); wprintw(mainwin,"%-5hu",ulist[usernum-1]->gid);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 4) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,11,2,"Ratio");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD); wprintw(mainwin,"%hu",ulist[usernum-1]->ratio);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 5) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,13,2,"Maximum connections");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD); wprintw(mainwin,"%i",ulist[usernum-1]->maxconn);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 6) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,15,2,"Home directory");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD);
   if(ulist[usernum-1]->homedir) wprintw(mainwin,"%s",ulist[usernum-1]->homedir);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 7) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,17,2,"Comment");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD);
   if(ulist[usernum-1]->comment) wprintw(mainwin,"%s",ulist[usernum-1]->comment);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 8) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,5,USERWINOFF,"Change password");
   wattroff(mainwin,A_REVERSE);

   if(fieldnum == 9) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,7,USERWINOFF,"Uploaded bytes");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD); wprintw(mainwin,"%qu",ulist[usernum-1]->ulbytes);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 10) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,9,USERWINOFF,"Downloaded bytes");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD); wprintw(mainwin,"%qu",ulist[usernum-1]->dlbytes);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 11) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,11,USERWINOFF,"Credits");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD); wprintw(mainwin,"%qu",ulist[usernum-1]->credits);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 12) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,13,USERWINOFF,"Flags");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   for(i=0;i<cfg.flagcount;i++)
   {
      if(ulist[usernum-1]->flags && strchr(ulist[usernum-1]->flags,cfg.flags[i].mark))
         wattron(mainwin,A_BOLD);
      waddch(mainwin,cfg.flags[i].mark); wattroff(mainwin,A_BOLD);
   }
}

void printgroup(int groupnum, int fieldnum)
{
   int i,j;

   for(i=2;i<GROUPWINH-2;i++)
   {
      wmove(mainwin,i,2);
      for(j=0;j<GROUPWINW-4;j++) waddch(mainwin,' ');
   }
   if(gnum) mvwprintw(mainwin,2,2,"Group %i of %i",groupnum,gnum);
   else
   {
      mvwprintw(mainwin,2,2,"Group 0 of 0"); return;
   }
   if(fieldnum == 1) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,5,2,"Name");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD); wprintw(mainwin,glist[groupnum-1]->name);
   wattroff(mainwin,A_BOLD);

   if(fieldnum == 2) wattron(mainwin,A_REVERSE);
   mvwprintw(mainwin,7,2,"ID");
   wattroff(mainwin,A_REVERSE); wprintw(mainwin,": ");
   wattron(mainwin,A_BOLD); wprintw(mainwin,"%-5hu ",glist[groupnum-1]->id);
   wattroff(mainwin,A_BOLD);
}

char editflags(int xpos, int ypos, unsigned char **flags)
{
   int i, pos = 0, ch, attr; unsigned char *newflags = NULL, *flpos;

   if(*flags)
   {
      newflags = malloc(strlen(*flags)+1); strcpy(newflags,*flags);
   }
   do
   {
      for(i=0;i<cfg.flagcount;i++)
      {
         attr = 0; if(i == pos) attr |= A_REVERSE;
         if(newflags && strchr(newflags,cfg.flags[i].mark)) attr |= A_BOLD;
         mvwaddch(mainwin,ypos,xpos+i,cfg.flags[i].mark|attr);
      }
      ch = wgetch(mainwin);
      if(ch == KEY_LEFT && pos > 0) pos--;
      else if(ch == KEY_RIGHT && pos < cfg.flagcount-1) pos++;
      else if(ch == 32)
      {
         if(newflags) flpos=strchr(newflags,cfg.flags[pos].mark);
         else flpos = NULL;
         if(flpos)
         {
            strcpy(flpos,flpos+1);
            newflags = realloc(newflags,strlen(newflags));
         }
         else
         {
            if(newflags) newflags = realloc(newflags,strlen(newflags)+2);
            else newflags = malloc(2);
            newflags[strlen(newflags)] = cfg.flags[pos].mark;
            newflags[strlen(newflags)+1] = 0;
         }
      }
   } while(ch != 10);
   if((newflags != *flags) || (newflags && *flags && strcmp(newflags,*flags)))
   {
      free(*flags); *flags = newflags; return 1;
   }
   free(newflags); return 0;
}

char editpassword(char **outpasswd)
{
   WINDOW *passwindow = NULL, *field1 = NULL, *field2 = NULL;
   time_t aika; char suola[2], f1str[21], f2str[21];
   int ch, pos = 0, retvalue = 0;

   time(&aika); srandom((unsigned int)aika);
   do
   {
      suola[0] = random();
   } while( !isalnum(suola[0]) );
   aika += 3; srandom((unsigned int)aika);
   do
   {
      suola[1] = random();
   } while( !isalnum(suola[1]) );
   
   passwindow = derwin(mainwin,11,25,USERWINH/2-5,USERWINW/2-12);
   wattrset(passwindow,COLOR_PAIR(4));
   drawwin(passwindow,25,11,"Change password",NULL);
   mvwprintw(passwindow,2,2,"New password");
   mvwprintw(passwindow,6,2,"Repeat");
   field1 = derwin(passwindow,1,20,4,2); keypad(field1,TRUE);
   field2 = derwin(passwindow,1,20,8,2); keypad(field2,TRUE);
   wattrset(field1,COLOR_PAIR(2)); wattrset(field2,COLOR_PAIR(2));
   curs_set(1); werase(field1); werase(field2); wmove(field1,0,0);
   touchwin(passwindow); touchwin(mainwin); wrefresh(passwindow);
   memset(f1str,0,21); memset(f2str,0,21);
   while((ch = wgetch(field1)) != 10 && ch != 27)
   {
      if(isprint(ch) && pos < 20)
      {
         wechochar(field1,'*'); f1str[pos++] = ch;
         if(pos == 20) wmove(field1,0,--pos);
      }
      else if((ch == KEY_BACKSPACE || ch == 127) && pos > 0)
      {
         mvwdelch(field1,0,--pos); f1str[pos] = 0;
      }
      else if((ch == KEY_DC || ch == 4) && pos < 20)
      {
         wdelch(field1); f1str[pos] = 0;
      }
   }
   pos = 0;
   if(ch != 27)
   {
      wmove(field2,0,0); 
      while((ch = wgetch(field2)) != 10 && ch != 27)
      {
         if(isprint(ch) && pos < 20)
         {
            wechochar(field2,'*'); f2str[pos++] = ch;
            if(pos == 20) wmove(field2,0,--pos);
         }
         else if((ch == KEY_BACKSPACE || ch == 127) && pos > 0)
         {
            mvwdelch(field2,0,--pos); f2str[pos] = 0;
         }
         else if((ch == KEY_DC || ch == 4) && pos < 20)
         {
            wdelch(field2); f2str[pos] = 0;
         }
      }
   }
   if(ch == 10)
   {
      retvalue = 1;
      if(strcmp(f1str,f2str))
      {
         notify("Passwords don't match!"); retvalue = 0;
      }
   }
   delwin(field1); delwin(field2); delwin(passwindow);
   if(retvalue)
   {
      for(ch=0;ch<strlen(f1str);ch++) if(f1str[ch] == 32) f1str[ch] = 0;
      *outpasswd = crypt(f1str,suola);
   }
   curs_set(0); return retvalue;
}

char editfield(int xpos, int ypos, int maxw, char *text, int numeric,
               char **outtext)
{
   WINDOW *editwindow = NULL; int ch, pos = 0;
   char tempstr[maxw+1];

   editwindow = derwin(mainwin,1,maxw,ypos,xpos);
   wattrset(editwindow,COLOR_PAIR(2)); werase(editwindow); curs_set(1);
   keypad(editwindow,TRUE); wmove(editwindow,0,0); touchwin(mainwin);
   if(text)
   {
      wprintw(editwindow,text); wrefresh(editwindow);
      pos = strlen(text);
   }
   while((ch = wgetch(editwindow)) != 10)
   {
      if(((!numeric && isprint(ch) && ch != ':') || 
       (numeric && isdigit(ch))) && pos < maxw)
      {
         wechochar(editwindow,ch); pos++;
         if(pos == maxw) wmove(editwindow,0,--pos);
      }
      else if((ch == KEY_BACKSPACE || ch == 127) && pos > 0)
         mvwdelch(editwindow,0,--pos);
      else if((ch == KEY_DC || ch == 4) && pos < maxw) wdelch(editwindow);
   }
   memset(tempstr,0,maxw+1);
   mvwinnstr(editwindow,0,0,tempstr,maxw);
   curs_set(0); delwin(editwindow);
   for(ch=0;ch < strlen(tempstr);ch++) if(tempstr[ch] == 32) tempstr[ch] = 0;
   if(strlen(tempstr))
   {
      *outtext = malloc(strlen(tempstr)+1); strcpy(*outtext,tempstr);
      if(!text || strcmp(tempstr,text)) return 1;
      else return 0;
   }
   *outtext = NULL;
   if(text) return 1;
   return 0;
}

/* Helper functions for uselect() */

unsigned int getnextuid(unsigned int uid)
{
   int i;

   uid++;
   for(i=0;i<unum;i++)
      if(ulist[i]->uid == uid)
      {
         i = -1; uid++;
      }
   return uid;
}

void deletemember(struct dftp_gent *group, char *name)
{
   int i,j;

   for(i=0;i<group->membercount;i++)
      if(!strcmp(group->members[i],name))
      {
         free(group->members[i]);
         for(j=i+1;j<group->membercount;j++)
            group->members[j-1] = group->members[j];
         group->members = realloc(group->members,
          --group->membercount*sizeof(void *)); i--;
      }
}

int changegid(struct dftp_uent *entry, unsigned int newgid)
{
   int i;

   for(i=0;i<gnum;i++) if(glist[i]->id == newgid) break;
   if(i == gnum)
   {
      notify("Specified group does not exist"); return 0;
   }
   addmember(glist[i],entry->name); /* Add user to the new group */
   for(i=0;i<gnum;i++) if(glist[i]->id == entry->gid) break;
   if(i < gnum) /* Then delete user from previous group */
      deletemember(glist[i],entry->name);
   entry->gid = newgid; return 1;
}

void renameuser(char *oldname, char *newname)
{
   int i,j;

   for(i=0;i<gnum;i++)
      for(j=0;j<glist[i]->membercount;j++)
         if(!strcmp(glist[i]->members[j],oldname))
         {
            free(glist[i]->members[j]);
            glist[i]->members[j] = malloc(strlen(newname)+1);
            strcpy(glist[i]->members[j],newname);
         }
   strcpy(oldname,newname);
}

void khelp()
{
   WINDOW *helpwin; int ch;

   helpwin = newwin(12,31,LINES/2-6,COLS/2-15);
   keypad(helpwin,TRUE); wattrset(helpwin,COLOR_PAIR(5)|A_BOLD);
   drawwin(helpwin,31,12,"Keyboard commands",NULL);
   mvwprintw(helpwin,2,2,"arrow keys    move around");
   mvwprintw(helpwin,3,2,"; / home      first entry");
   mvwprintw(helpwin,4,2,"n / insert    new entry");
   mvwprintw(helpwin,5,2,"d / delete    delete entry");
   mvwprintw(helpwin,6,2,", / pageup    previous entry");
   mvwprintw(helpwin,7,2,". / pagedown  next entry");
   mvwprintw(helpwin,8,2,"; / home      first entry");
   mvwprintw(helpwin,9,2,": / end       last entry");    

   do
   {
      ch = wgetch(helpwin);
   } while(ch != KEY_F(1) && ch != 27 && ch != 'q' && ch != 10);
   werase(helpwin); wrefresh(helpwin); delwin(helpwin);
   if(mainwin)
   {
      touchwin(mainwin); wrefresh(mainwin);
   }
}

int uselect()
{
   int i,usernum=1,fieldnum=1,update=0;
   char *tempstr = NULL, numstr[16];

   init_mainwin(USERWINW,USERWINH);
   for(i=0;i != 'q' && i != 27;)
   {
      printuser(usernum,fieldnum); wrefresh(mainwin);
      i = wgetch(mainwin);
      if(i == KEY_F(1) || i == 'h') khelp();
      else if(i == KEY_IC || i == 'n')
      {
         if(!gnum) notify("Create a group first!");
         else
         {
            ulist = realloc(ulist,++unum*sizeof(void *));
            ulist[unum-1] = calloc(1,sizeof(struct dftp_uent));
            strcpy(ulist[unum-1]->passwd,"*");
            if(unum > 1)
            {
               ulist[unum-1]->uid = getnextuid(ulist[usernum-1]->uid);
               ulist[unum-1]->gid = ulist[usernum-1]->gid;
               ulist[unum-1]->maxconn = ulist[usernum-1]->maxconn;
            }
            else
            {
               ulist[unum-1]->uid = FIRSTUID;
               ulist[unum-1]->gid = glist[gnum-1]->id;
               ulist[unum-1]->maxconn = 1;
            }
            usernum = unum; update = 3;
         }
      }
      if(unum)
      {
         if((i == KEY_PPAGE || i == ',') && usernum > 1) usernum--;
         else if((i == KEY_NPAGE || i == '.') && usernum < unum) usernum++;
         else if(i == KEY_HOME || i == ';') usernum = 1;
         else if(i == KEY_END || i == ':') usernum = unum;
         else if(i == KEY_RIGHT && fieldnum+7 <= USERFIELDS) fieldnum += 7;
         else if(i == KEY_LEFT && fieldnum-7 > 0) fieldnum -= 7;
         else if(i == KEY_UP && fieldnum > 1) fieldnum--;
         else if(i == KEY_DOWN && fieldnum < USERFIELDS) fieldnum++;
         else if((i == KEY_DC || i == 4 || i == 'd') &&
          confirm("Really delete user?",0))
         {
            int j;

            for(j=0;j<gnum;j++) deletemember(glist[j],ulist[usernum-1]->name);
            dftp_ufree(ulist[usernum-1]);
            for(j=usernum-1;j<unum-1;j++) ulist[j] = ulist[j+1];
            ulist = realloc(ulist,--unum*sizeof(void *));
            if(usernum > unum && usernum > 1) usernum--;
            update = 3;
         }
         else if(i == 10)
            switch(fieldnum)
            {
               case 1: if(editfield(8,5,16,ulist[usernum-1]->name,0,&tempstr))
                       {
                          renameuser(ulist[usernum-1]->name,tempstr);
                          update = 3;
                       }
                       free(tempstr); break;
               case 2: sprintf(numstr,"%hu",ulist[usernum-1]->uid);
                       if(editfield(7,7,5,numstr,1,&tempstr))
                       {
                          sscanf(tempstr,"%u",&ulist[usernum-1]->uid);
                          update = 1;
                       }
                       free(tempstr); break;
               case 3: sprintf(numstr,"%hu",ulist[usernum-1]->gid);
                       if(editfield(7,9,5,numstr,1,&tempstr))
                       {
                          unsigned int tempgid;

                          sscanf(tempstr,"%u",&tempgid);
                          if(changegid(ulist[usernum-1],tempgid)) update = 3;
                       }
                       free(tempstr); break;
               case 4: sprintf(numstr,"%hu",ulist[usernum-1]->ratio);
                       if(editfield(9,11,2,numstr,1,&tempstr))
                       {
                          sscanf(tempstr,"%hu",&ulist[usernum-1]->ratio);
                          update = 1;
                       }
                       free(tempstr); break;
               case 5: sprintf(numstr,"%hu",ulist[usernum-1]->maxconn);
                       if(editfield(23,13,3,numstr,1,&tempstr))
                       {
                         sscanf(tempstr,"%hu",&ulist[usernum-1]->maxconn);
                         update = 1;
                       }
                       free(tempstr); break;
               case 6: if(editfield(18,15,USERWINW-22,ulist[usernum-1]->homedir,0,&ulist[usernum-1]->homedir))
                          update = 1; break;
               case 7: if(editfield(11,17,USERWINW-15,ulist[usernum-1]->comment,0,&ulist[usernum-1]->comment))
                          update = 1; break;
               case 8: if(editpassword(&tempstr))
                       {
                          strcpy(ulist[usernum-1]->passwd,tempstr);
                          update = 1;
                       }
                       break;
               case 9: sprintf(numstr,"%qu",ulist[usernum-1]->ulbytes);
                       if(editfield(USERWINOFF+16,7,14,numstr,1,&tempstr))
                       {
                          sscanf(tempstr,"%qu",&ulist[usernum-1]->ulbytes);
                          update = 1;
                       }
                       free(tempstr); break;
               case 10: sprintf(numstr,"%qu",ulist[usernum-1]->dlbytes);
                       if(editfield(USERWINOFF+18,9,14,numstr,1,&tempstr))
                       {
                          sscanf(tempstr,"%qu",&ulist[usernum-1]->dlbytes);
                          update = 1;
                       }
                       free(tempstr); break;
               case 11: sprintf(numstr,"%qu",ulist[usernum-1]->credits);
                       if(editfield(USERWINOFF+9,11,14,numstr,1,&tempstr))
                       {
                          sscanf(tempstr,"%qu",&ulist[usernum-1]->credits);
                          update = 1;
                       }
                       free(tempstr); break;
               case 12: if(editflags(USERWINOFF+7,13,&ulist[usernum-1]->flags))
                          update = 1; break;
            }
      }
   }
   werase(mainwin); wrefresh(mainwin); delwin(mainwin); mainwin = NULL;
   return update;
}

/* Helper functions for gselect() */

unsigned int getnextid(unsigned int id) /* Finds next available group id */
{
   int i;

   id++;
   for(i=0;i<gnum;i++)
      if(glist[i]->id == id)
      {
         i = -1; id++;
      }
   return id;
}

int gselect()
{
   int i,groupnum=1,fieldnum=1,update=0;
   char *tempstr = NULL, numstr[16];

   init_mainwin(GROUPWINW,GROUPWINH);
   for(i=0;i != 'q' && i != 27;)
   {
      printgroup(groupnum,fieldnum); wrefresh(mainwin);
      i = wgetch(mainwin);
      if(i == KEY_F(1) || i == 'h') khelp();
      else if(i == KEY_IC || i == 'n')
      {
         glist = realloc(glist,++gnum*sizeof(void *));
         glist[gnum-1] = calloc(1,sizeof(struct dftp_gent));
         if(gnum > 1) glist[gnum-1]->id = getnextid(glist[groupnum-1]->id);
         else glist[gnum-1]->id = FIRSTGID;
         groupnum = gnum; update = 2;
      }
      else if(groupnum)
      {
         if(i == KEY_PPAGE && groupnum > 1) groupnum--;
         else if(i == KEY_NPAGE && groupnum < gnum) groupnum++;
         else if(i == KEY_HOME) groupnum = 1;
         else if(i == KEY_END) groupnum = gnum;
         else if(i == KEY_UP && fieldnum > 1) fieldnum--;
         else if(i == KEY_DOWN && fieldnum < GROUPFIELDS) fieldnum++;
         else if((i == KEY_DC || i == 4 || i == 'd') &&
          confirm("Really delete group?",0))
         {
            int j;

            for(j=0;j<unum;j++)
               if(ulist[j]->gid == glist[groupnum-1]->id)
               {
                  notify("Unable to delete group - it is the primary group for user #%i",j+1);
                  break;
               }
            if(j == unum)
            {
               dftp_gfree(glist[groupnum-1]);
               for(j=groupnum-1;j<gnum-1;j++) glist[j] = glist[j+1];
               glist = realloc(glist,--gnum*sizeof(void *));
               if(groupnum > gnum && groupnum > 1) groupnum--; update = 2;
            }
         }
         else if(i == 10)
            switch(fieldnum)
            {
               case 1: if(editfield(8,5,16,glist[groupnum-1]->name,0,&tempstr))
                       {
                          strcpy(glist[groupnum-1]->name,tempstr); update = 2;
                       }
                       free(tempstr); break;
               case 2: sprintf(numstr,"%hu",glist[groupnum-1]->id);
                       if(editfield(6,7,5,numstr,1,&tempstr))
                       {
                          unsigned int newid;

                          sscanf(tempstr,"%u",&newid);
                          for(i=0;i<unum;i++)
                             if(ulist[i]->gid == glist[groupnum-1]->id)
                                ulist[i]->gid = newid;
                          glist[groupnum-1]->id = newid;
                          update = 3;
                       }
                       free(tempstr); break;
            }
      }
   }
   werase(mainwin); wrefresh(mainwin); delwin(mainwin); mainwin = NULL;
   return update;
}

int quit(int update)
{
   if(update && confirm("Save changes?",0))
   {
      if(!validateusers()) return 1;
      if(update&1 && !save_users())
         notify("Error writing user database");
      if(update&2 && !save_groups())
         notify("Error writing group database");
   }
   return 0;
}

int menu()
{
   WINDOW *akkuna; int i; static int pos = 0, update = 0;

   akkuna = newwin(MENUH,MENUW,LINES/2-MENUH/2,COLS/2-MENUW/2);
   wattrset(akkuna,COLOR_PAIR(1)); keypad(akkuna,TRUE);
   werase(akkuna);
   if(has_colors()) wborder(akkuna,0,0,0,0,0,0,0,0);
   else wborder(akkuna,'|','|','-','-','+','+','+','+');
   for(i=0;;)
   {
      menuitem(akkuna,"Modify user database",0,pos);
      menuitem(akkuna,"Modify group database",1,pos);
      menuitem(akkuna,"Rebuild user&group databases",2,pos);
      menuitem(akkuna,"Exit",3,pos);
      wrefresh(akkuna);
      i = wgetch(akkuna);
      if(i == KEY_UP && pos > 0) pos--;
      else if(i == KEY_DOWN && pos < MENUITEMS-1) pos++;
      else if(i == 'q' || i == 27)
      {
         delwin(akkuna); return quit(update);
      }
      else if(i == 10)
      {
         switch(pos)
         {
            case 0: update |= uselect(); delwin(akkuna); return 1;
            case 1: update |= gselect(); delwin(akkuna); return 1;
            case 2: rebuild(); notify("Rebuild complete");
                    delwin(akkuna); return 1;
            case 3: delwin(akkuna); return quit(update);
         }
      }      
   }
   return 0;
}

int main(int argc, char *argv[])
{
   if(argc > 1 && (!strcmp(argv[1],"--help") || !strcmp(argv[1],"-h"))) usage();

   dftp_readconfig(&cfg);
   passwdname = malloc(strlen(cfg.confdir)+strlen(USERFILE)+2);
   sprintf(passwdname,"%s/%s",cfg.confdir,USERFILE);
   sitepasswdname = malloc(strlen(cfg.sitedir)+strlen("passwd")+6);
   sprintf(sitepasswdname,"%s/etc/passwd",cfg.sitedir);
   groupname = malloc(strlen(cfg.confdir)+strlen(GROUPFILE)+2);
   sprintf(groupname,"%s/%s",cfg.confdir,GROUPFILE);
   sitegroupname = malloc(strlen(cfg.sitedir)+strlen("group")+6);
   sprintf(sitegroupname,"%s/etc/group",cfg.sitedir);

   if(argc > 1)
   {
      if(!strcmp(argv[1],"rebuild")) rebuild();
      else if(!strcmp(argv[1],"ulist")) listusers(argv[2]);
      else if(!strcmp(argv[1],"glist")) listgroups(argv[2]);
      else fprintf(stderr,"Unknown command '%s'\n",argv[1]);
   }
   else
   {
      visual_setup();
      load_users();  load_groups();
      while(menu());
      clear(); refresh(); endwin();
   }
   return 0;
}
