#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <fcntl.h>
#include <time.h>
#include <fnmatch.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <signal.h>
#include "dftpd.h"

/* libdftp dependencies */

int dftp_ugetbyname(int, char *, struct dftp_uent *);
int dftp_ugetbyid(int, int, struct dftp_uent *);
void dftp_uupdatedb(int, struct dftp_uent *);
int dftp_dupeaddent(int, struct dftp_dupeent *);
int dftp_dupegetent(int, struct dftp_dupeent *);
char *dftp_printbytes(unsigned long long);

/* dftpd.c dependencies */

void ftplog(const char *, const char *, ...);
void activity(const char *);
int open_dataconn(char, char, char *);
void close_dataconn(int);

extern struct dftp_config cfg;
extern int sockfd, passwdfile, dupefile;
extern struct dftp_vars var;

/* aux.c dependencies */

int checklog();
int checkaccess(char);
char **parsestr(char *);
void _status(unsigned short);

/* srv-common.c dependencies */

void writes(const char *,...);

/* cmd.c dependencies */

extern unsigned long offset;

/* private variables */

int filefd = -1, datasock = -1;
unsigned long totwritten = 0, nwritten = 0, fsize = 0;
char *fname = NULL;

void chbytes(unsigned long long *bytet, long long maara)
{
   if(maara < 0 && *bytet < -maara) *bytet = 0;
   else *bytet += maara;
}

/* Download related functions */

int check_credits()
{
   if(!var.user.ratio || var.userclass == REAL) return 1;
   if(var.user.credits > fsize)
   {
      writes("430 Insufficient credits to download file.\r\n");
      close(filefd); return 0;
   }
   return 1;
}

void deduct_credits()
{
   if(var.userclass == REAL || !var.user.ratio) return;
   if(var.userclass == ANONYMOUS)
   {
      if(var.user.ratio > 0) chbytes(&var.user.credits, -(long long)fsize);
      return;
   }
   flock(passwdfile,LOCK_EX);
   if(!dftp_ugetbyid(passwdfile,var.user.uid,&var.user))
   {
      ftplog("ERR","Panic! Current user not found in user database!");
      writes("535 Current user entry not found in user database - closing control connection.\r\n");
      exit(1);
   }
   chbytes(&var.user.credits, -(long long)fsize);
   dftp_uupdatedb(passwdfile, &var.user);
   flock(passwdfile,LOCK_UN);
}

void compensate_credits()
{
   if(var.userclass == REAL) return;

   flock(passwdfile,LOCK_EX);
   if(dftp_ugetbyname(passwdfile,var.user.name,&var.user) >= 0)
   {
      ftplog("ERR","Panic! Current user not found in user database!");
      writes("535 Current user entry not found in user database - closing control connection.");
      exit(1);
  }
  chbytes(&var.sessiondl,totwritten);
  chbytes(&var.user.dlbytes,totwritten);
  if(var.user.ratio > 0 && var.userclass == GUEST)
     chbytes(&var.user.credits,fsize-totwritten);
  dftp_uupdatedb(passwdfile, &var.user);
  flock(passwdfile,LOCK_UN);
  if(var.user.ratio > 0 && var.userclass != GUEST)
     chbytes(&var.user.credits,fsize-totwritten);
}

void start_statusupdater();

void status_update(int sig)
{
   char *event; static unsigned int lastvalue = 0;

   if(sig != SIGALRM) return;

   event = malloc(strlen(fname)+50);
   if(fsize) sprintf(event,"RETR %s - %.0f%% of %s (%.1fKb/s)",fname,(float)totwritten/(float)fsize*100.0,
    dftp_printbytes(fsize),(float)(totwritten-lastvalue)/1024.0/(float)cfg.stupdatetime);
   else sprintf(event,"RETR %s - %s",fname,dftp_printbytes(totwritten));
   activity(event); signal(SIGVTALRM,status_update); free(event);
   start_statusupdater();
}

void start_statusupdater()
{
   struct itimerval value;

   value.it_value.tv_sec = cfg.stupdatetime;
   value.it_value.tv_usec = 0;
   value.it_interval.tv_sec = 0;
   value.it_interval.tv_usec = 0;
   signal(SIGALRM,status_update);
   setitimer(ITIMER_REAL,&value,NULL);
}

void stop_statusupdater()
{
   static unsigned int oldvalue = 0; struct itimerval value;

   if(totwritten == oldvalue) return;
   memset(&value,0,sizeof(struct itimerval));
   signal(SIGVTALRM,SIG_IGN);
   setitimer(ITIMER_VIRTUAL,&value,NULL);
}

void send_binary()
{
   char buf[10*KB], c[6], *bufptr; unsigned long nleft;

   while((nleft = read(filefd,&buf,sizeof(buf))) > 0)
   {
      bufptr = buf;
      while(nleft > 0)
      {
         nwritten = write(datasock, bufptr, nleft);
         if(nwritten <= 0) break;
         
         nleft -= nwritten;
         bufptr += nwritten; totwritten += nwritten;
         if( read(sockfd,c,1) > 0)
         {
            if(read(sockfd,c,6) > 0 && !strncasecmp(c,"ABOR",4))
            {
               close_dataconn(datasock); datasock = -1;
               writes("426 Transfer aborted. Data connection closed.\r\n");
               writes("226 Abort successful.\r\n"); return;
            }
         }
      }
      if(nwritten <= 0) break;
   }
}

void send_ascii()
{
   char buf[10*KB], c[6], *bufptr;
   unsigned long nleft;

   while(read(filefd,&buf,1) > 0)
   {
      nleft = 1; bufptr = buf;
      if(buf[0] == '\n')
      {
         buf[0] = '\r'; buf[1] = '\n'; nleft = 2;
      }
      while(nleft > 0)
      {
         nwritten = write(datasock, bufptr, nleft);
         if( nwritten < 0) break;

         nleft -= nwritten;
         bufptr += nwritten; totwritten += nwritten;
         if( read(sockfd,c,1) > 0)
         {
            read(sockfd,c,6);
            if(!strncasecmp(c,"ABOR",4))
            {
               close_dataconn(datasock); datasock = -1;
               writes("426 Transfer aborted. Data connection closed.\r\n");
               writes("226 Abort successful.\r\n"); return;
            }
         }
      }
      if(nwritten < 0) break;
   }
}

void get(char *args)
{
   char textbuf[NAME_MAX+30]; struct stat statbuf;
   double timediff; time_t starttime, endtime;

   if(!checklog()) return;
   if(var.userclass != REAL && !checkaccess('R')) return;

   filefd = open(args,O_RDONLY);
   if(filefd < 0)
   {
      writes("550 %s: %s.\r\n",args,strerror(errno));
      return;
   }

   fstat(filefd,&statbuf);
   fsize = statbuf.st_size; fname = args;
   if(!check_credits()) return;
   deduct_credits();

   if(offset > 0) lseek(filefd, offset, SEEK_SET);
   snprintf(textbuf,sizeof(textbuf),"%s (%lu bytes)",args,fsize);
   datasock = open_dataconn(var.trans_type,ID_DATA,textbuf);

   fcntl(sockfd,F_SETFL,O_NONBLOCK);
   start_statusupdater(); time(&starttime);
   if(var.trans_type == 'A') send_ascii();
   else send_binary();
   time(&endtime); stop_statusupdater();
   close(filefd); offset = 0; close_dataconn(datasock);
   fcntl(sockfd,F_SETFL,0);

   /* Update user downloaded bytes status and compensate credits */

   compensate_credits();

   if(nwritten < 0) writes("426 Error writing to data socket.\r\n");
   else
   {
      timediff = difftime(endtime,starttime);
      if(totwritten >= fsize)
      {
         _status(226);
         ftplog("SEND","Sent %s (%u bytes) in %.0f seconds",args,totwritten,timediff);
      }
      else ftplog("SEND","Sent %s (%u bytes) in %.0f seconds - transfer aborted",args,totwritten,timediff);
   }
}

/* Upload related functions */

int check_ban(char *file)
{
   int i;

   for(i=0;i<cfg.uccount;i++)
   {
      if(*cfg.ulchecker[i].handler == '-' && !fnmatch(cfg.ulchecker[i].mask,file,FNM_PERIOD))
      {
         writes("533 Upload request denied -- file match banned filemask '%s'.\r\n",cfg.ulchecker[i].mask);
         return 0;
      }
   }
   return 1;
}

int check_dupe(char *file)
{
   int i, dupe = 1; struct dftp_dupeent dentry; struct tm *aika2;

   for(i=0;i<cfg.dccount;i++)
      if(!fnmatch(cfg.dupechecker[i],file,FNM_PERIOD))
      {
         dupe = 2; flock(dupefile,LOCK_EX); lseek(dupefile,0,SEEK_SET);
         while(dftp_dupegetent(dupefile,&dentry))
         {
            free(dentry.dirname);
            if(!strcmp(file,dentry.filename))
            {
               aika2 = localtime(&dentry.ultime);
               writes("550 '%s' already exists, uploaded by %s on %02i:%02i %i.%i.%i\r\n"
                ,file,dentry.uploader,aika2->tm_hour,aika2->tm_min,aika2->tm_mday,aika2->tm_mon,1900+aika2->tm_year);
               free(dentry.filename); flock(dupefile,LOCK_UN); return 0;
            }
            free(dentry.filename);
         }
         flock(dupefile,LOCK_UN);
      }
   return dupe;
}

void add_dupe(char *file, time_t endtime)
{
   struct dftp_dupeent dentry; char buf[PATH_MAX+1], *ptr;

   if(cfg.debuglevel >= 2) ftplog("RECV","DEBUG: creating dupe entry");
   getcwd(buf,PATH_MAX);
   ptr = strrchr(buf,'/'); *ptr = 0;
   dentry.dirname = malloc(strlen(buf)+1);
   dentry.filename = malloc(strlen(file)+1);
   strcpy(dentry.filename,file);
   strcpy(dentry.dirname,buf);
   dentry.ultime = endtime;
   strcpy(dentry.uploader,var.user.name);
   flock(dupefile,LOCK_EX); lseek(dupefile,0,SEEK_END);
   dftp_dupeaddent(dupefile,&dentry);
   flock(dupefile,LOCK_UN);
   free(dentry.filename);
   free(dentry.dirname);
}

int ulcheck(char *filename)
{
   int result = 0,status = 0,cpid,i; char *cmdline = NULL;

   if(cfg.uccount == 0) return 0;
   if(cfg.debuglevel >= 2) ftplog("RECV","DEBUG: starting upload check");
   for(i = 0; status == 0 && i < cfg.uccount; i++)
   {
      if(!fnmatch(cfg.ulchecker[i].mask,filename,FNM_PERIOD))
      {
         if(*cfg.ulchecker[i].handler == '+')
         {
            writes("226-File accepted.\r\n"); return -1;
         }
         else
         {
            if(access(cfg.ulchecker[i].handler,X_OK))
            {
               writes("226-Cannot find/execute %s.\r\n",cfg.ulchecker[i].handler);
               return 0;
            }
            cmdline = malloc(strlen(cfg.ulchecker[i].handler)+1);
            strcpy(cmdline,cfg.ulchecker[i].handler);
            status = 1;
         }
      }
   }
   if(status == 0)
   {
      if(cfg.debuglevel >= 1)
      {
         writes("226-Unknown file type");
         if(var.user.ratio) writes(" - no credits awarded"); writes("\r\n");
      }
      return 0;
   }

   signal(SIGCLD,SIG_DFL);
   if((cpid = fork()) < 0) return 0;
   if(cpid == 0)
   {
      struct stat statbuf; char **cmd = NULL, text[20];

      close(sockfd); for(i=0;i<20;i++) close(i);

      setenv("FILENAME",filename,1); stat(filename,&statbuf);
      sprintf(text,"%lu",statbuf.st_size); setenv("FILESIZE",text,1);
      cmd = parsestr(cmdline); execv(cmd[0],cmd); exit(99);
   }
   wait(&status); signal(SIGCLD,SIG_IGN);
   result = WEXITSTATUS(status);
   if(cfg.debuglevel >= 1) ftplog("RECV","DEBUG: %s exited with code %i",cmdline,result);
   if(result)
   {
      writes("226-File '%s' FAILED upload check!\r\n",filename);
      return 0;
   }
   writes("226-Upload check successful.\r\n");
   return -1;
}

void put(char *args)
{
   char buf[10*KB]; int dupe = 0; unsigned long nleft;
   double timediff; time_t starttime, endtime;

   if(!checklog()) return;
   if(var.userclass != REAL && !checkaccess('W')) return;

   if(!check_ban(args)) return;
   if(!(dupe = check_dupe(args))) return;
   if(!access(args,F_OK) && (!var.user.flags || !strchr(var.user.flags,'D')))
   {
      writes("532 File '%s' already exists; delete/rename/overwrite access denied.\r\n",args);
      return;
   }

   filefd = open(args,O_CREAT|O_WRONLY,var.cmode);
   if(filefd < 0)
   {
      writes("550 %s: %s.\r\n",args,strerror(errno));
      return;
   }
   if(offset > 0) { lseek(filefd,offset,SEEK_SET); offset = 0; }
   else ftruncate(filefd,0); fname = args;

   datasock = open_dataconn(var.trans_type,ID_DATA,args);

   start_statusupdater(); time(&starttime);
   while( (nleft = read(datasock,&buf,sizeof(buf))) > 0) 
   {
      nwritten = write(filefd, &buf, nleft);
      if(nwritten < 0) break;
      totwritten += nwritten;
   }
   time(&endtime); stop_statusupdater();
   close_dataconn(datasock); close(filefd);
   if(nleft < 0 || nwritten < 0)
   {
      ftplog("ERR","Error occurred during receive: %s",strerror(errno));
      writes("426 Transfer aborted\r\n"); return;
   }
   timediff = difftime(endtime,starttime);

   if(var.userclass == REAL || ulcheck(args))
   {
      ftplog("RECV","Received %s (%u bytes) in %.0f seconds",args,totwritten,timediff);
      chbytes(&var.sessionul,totwritten);

      /* Update user uploaded bytes and credit status */

      if(var.userclass != REAL)
      {
         flock(passwdfile,LOCK_EX);
         if(!dftp_ugetbyid(passwdfile,var.user.uid,&var.user))
         {
            flock(passwdfile,LOCK_UN);
            ftplog("ERR","Panic! Current user not found in user database!");
            writes("535 Current user entry not found in user database - closing control connection.");
            exit(1);
         }
         chbytes(&var.user.ulbytes,totwritten);
         if(var.userclass != ANONYMOUS)
            chbytes(&var.user.credits,totwritten*var.user.ratio);
         dftp_uupdatedb(passwdfile, &var.user);
         flock(passwdfile,LOCK_UN);
         if(var.userclass == ANONYMOUS)
            chbytes(&var.user.credits,totwritten*var.user.ratio);
         if(dupe == 2) add_dupe(args, endtime);
      }
   }
   else ftplog("RECV","Received %s (%u bytes) in %.0f seconds - upload check failed",args,totwritten,timediff);
   _status(226);
}
